/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "quakedef.h"
#include "gl_mtex.h"

extern	model_t	*loadmodel;
extern cvar_t r_detailsky;

int		skytexturenum;
int		gl_useskybox;

int		solidskytexture;
int		alphaskytexture;
int		simpleskytexture;
float	speedscale;		// for top sky and bottom sky

extern int global_worldtype;

glmsurface_t	*warpface;

extern cvar_t gl_subdivide_size;

cvar_t r_waterwarp = {"r_waterwarp", "0", true};

void BoundPoly (int numverts, float *verts, vec3_t mins, vec3_t maxs)
{
	int		i, j;
	float	*v;

	mins[0] = mins[1] = mins[2] = 9999;
	maxs[0] = maxs[1] = maxs[2] = -9999;

	v = verts;

	for (i = 0; i < numverts; i++)
	{
		for (j = 0; j < 3; j++, v++)
		{
			if (*v < mins[j]) mins[j] = *v;
			if (*v > maxs[j]) maxs[j] = *v;
		}
	}
}

void SubdividePolygon (int numverts, float *verts, float subdiv_size)
{
	int		i, j, k;
	vec3_t	mins, maxs;
	float	m;
	float	*v;
	vec3_t	front[64], back[64];
	int		f, b;
	float	dist[64];
	float	frac;
	glpoly_t	*poly;
	float	s, t;

	// can't really do anything here
	if (numverts > 60) Host_Error ("numverts = %i", numverts);

	BoundPoly (numverts, verts, mins, maxs);

	for (i = 0; i < 3; i++)
	{
		m = (mins[i] + maxs[i]) * 0.5;
		m = subdiv_size * floor (m / subdiv_size + 0.5);

		if (maxs[i] - m < 8) continue;
		if (m - mins[i] < 8) continue;

		// cut it
		v = verts + i;

		for (j = 0; j < numverts; j++, v += 3) dist[j] = *v - m;

		// wrap cases
		dist[j] = dist[0];
		v -= i;
		VectorCopy (verts, v);

		f = b = 0;
		v = verts;

		for (j = 0; j < numverts; j++, v += 3)
		{
			if (dist[j] >= 0) {VectorCopy (v, front[f]); f++;}
			if (dist[j] <= 0) {VectorCopy (v, back[b]); b++;}
			if (dist[j] == 0 || dist[j+1] == 0) continue;

			if ((dist[j] > 0) != (dist[j + 1] > 0))
			{
				// clip point
				frac = dist[j] / (dist[j] - dist[j + 1]);

				for (k = 0; k < 3; k++)
					front[f][k] = back[b][k] = v[k] + frac * (v[3 + k] - v[k]);

				f++;
				b++;
			}
		}

		SubdividePolygon (f, front[0], subdiv_size);
		SubdividePolygon (b, back[0], subdiv_size);
		return;
	}

	poly = Hunk_Alloc (sizeof (glpoly_t) + (numverts - 4) * VERTEXSIZE * sizeof (float));
	poly->next = warpface->warp_polys;
	warpface->warp_polys = poly;
	poly->numverts = numverts;

	for (i = 0; i < numverts; i++, verts += 3)
	{
		VectorCopy (verts, poly->verts[i]);
		s = DotProduct (verts, warpface->texinfo->vecs[0]);
		t = DotProduct (verts, warpface->texinfo->vecs[1]);

		// mh - speed up water rendering
		if (subdiv_size == 64)
		{
			s = s / 64;
			t = t / 64;
		}

		poly->verts[i][3] = s;
		poly->verts[i][4] = t;
		poly->verts[i][5] = 0;
	}
}

/*
================
GL_SubdivideSurface

Breaks a polygon up along axial 64 unit
boundaries so that turbulent and sky warps
can be done reasonably.

subdiv_size is 256 for sky and 64 for water
================
*/
void GL_SubdivideSurface (glmsurface_t *fa, int subdiv_size)
{
	vec3_t		verts[64];
	int			numverts;
	int			i;
	int			lindex;
	float		*vec;

	warpface = fa;

	// convert edges back to a normal polygon
	numverts = 0;

	for (i=0 ; i<fa->numedges ; i++)
	{
		lindex = loadmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->vertexes[loadmodel->edges[lindex].v[0]].position;
		else vec = loadmodel->vertexes[loadmodel->edges[-lindex].v[1]].position;

		VectorCopy (vec, verts[numverts]);
		numverts++;
	}

	SubdividePolygon (numverts, verts[0], (float) subdiv_size);
}

//=========================================================



/*
=============
EmitWaterPolys

Does a water warp on the pre-fragmented glpoly_t chain
=============
*/

// makes a surface into horrid wobbling jelly.  good for blocks of flesh, etc
void MakePervy (float *s, float *t, float wobble_speed)
{
	float os, ot, as, at;

	os = *s;
	ot = *t;

	as = os + sin((int)((ot*0.125+realtime) * wobble_speed) & 255);
	at = ot + sin((int)((os*0.125+realtime) * wobble_speed) & 255);

	*s = as * 0.015625;
	*t = at * 0.015625;
}


extern float s_mody;
extern float t_mody;


// speed up sin calculations - Ed
float turbsin[] =
{
	#include "gl_warp_sin.h"
};

#define TURBSCALE  40.74366543153  //(256.0 / (2 * M_PI))


// determine the s_mody and t_mody for static warps - this is only done once per frame
// moved to gl_warp.c from gl_rsurf.c so i can move turbsin and turbscale back in here.
void Calc_WarpModys (void)
{
	s_mody = turbsin[(int)(realtime * TURBSCALE) & 255] / 64.0;

	// previous versions used cl.time here (copied from lordhavoc) - this meant that t_mody
	// wouldn't update if the game was paused, in the console or in a menu.
	t_mody = turbsin[(int)(realtime * TURBSCALE + 96.0) & 255] / 64.0;
}


// consolidated all these functions into one.
void EmitWaterPolys (glmsurface_t *fa, float temp_wave, float trans_lev, float warp_factor)
{
	glpoly_t	*p;
	float		*v;
	int			i;
	float		nv;
	float		s, t, os, ot;


	for (p = fa->warp_polys; p; p = p->next)
	{
		glBegin (GL_POLYGON);

		for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
		{
			if (temp_wave)
				nv = v[2] + temp_wave + (temp_wave * sin (v[0] * 0.02 + realtime)
					* sin (v[1] * 0.02 + realtime) * sin (v[2] * 0.02 + realtime));
			else nv = v[2];

			if (warp_factor)
			{
				os = v[3] * 64;
				ot = v[4] * 64;

				s = os + turbsin[(int)((ot * 0.125 + realtime) * TURBSCALE) & 255];
				s *= (1.0 / 64);

				t = ot + turbsin[(int)((os * 0.125 + realtime) * TURBSCALE) & 255];
				t *= (1.0 / 64);

				glTexCoord2f (s, t);
			}
			else glTexCoord2f (v[3] + s_mody, v[4] + t_mody);

			glColor4f (v[8], v[9], v[10], trans_lev);
			glVertex3f (v[0], v[1], nv);
		}

		glEnd ();
	}
}


/*
=============
EmitSkyPolys
=============
*/
void EmitSkyPolys (glmsurface_t *fa)
{
	glpoly_t	*p;
	float		*v;
	int			i;
	float	s, t;
	vec3_t	dir;
	float	length;

	for (p=fa->warp_polys ; p ; p=p->next)
	{
		glBegin (GL_POLYGON);

		for (i=0,v=p->verts[0] ; i<p->numverts ; i++, v+=VERTEXSIZE)
		{
			VectorSubtract (v, r_origin, dir);
			dir[2] *= 3;	// flatten the sphere

			length = dir[0]*dir[0] + dir[1]*dir[1] + dir[2]*dir[2];
			length = sqrt (length);
			length = 6*63/length;

			dir[0] *= length;
			dir[1] *= length;

			s = (speedscale + dir[0]) * (1.0/128);
			t = (speedscale + dir[1]) * (1.0/128);

			glTexCoord2f (s, t);
			glVertex3fv (v);
		}
		glEnd ();
	}
}


// simple non-detailed sky drawing
void EmitSimpleSkyPolys (glmsurface_t *fa)
{
	glpoly_t *p;
	float *v;
	int i;

	for (p=fa->warp_polys ; p ; p=p->next)
	{
		glBegin (GL_POLYGON);

		for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
		{
			glTexCoord2fv (&v[0]);
			glVertex3fv (v);
		}

		glEnd ();
	}
}


/*
===============
EmitBothSkyLayers

Does a sky warp on the pre-fragmented glpoly_t chain
This will be called for brushmodels, the world
will have them chained together.
===============
*/
void EmitBothSkyLayers (glmsurface_t *fa)
{
	if (r_detailsky.value)
	{
		// slower moving 2 component sky
		glBindTexture (GL_TEXTURE_2D, solidskytexture);
		speedscale = realtime*8;
		speedscale -= (int)speedscale & ~127 ;

		EmitSkyPolys (fa);

		glEnable (GL_BLEND);
		glBindTexture (GL_TEXTURE_2D, alphaskytexture);
		oldtexture = alphaskytexture;
		speedscale = realtime*16;
		speedscale -= (int)speedscale & ~127 ;

		EmitSkyPolys (fa);

		glDisable (GL_BLEND);
		return;
	}

	// faster simple one-colour sky
	glBindTexture (GL_TEXTURE_2D, simpleskytexture);
	oldtexture = simpleskytexture;

	EmitSimpleSkyPolys (fa);
}


/*
=================
R_DrawSkyChain
=================
*/
void R_DrawSkyChain (glmsurface_t *s)
{
	glmsurface_t *fa;

	if (r_detailsky.value)
	{
		// slower moving 2 component sky
		glBindTexture (GL_TEXTURE_2D, solidskytexture);
		speedscale = realtime*8;
		speedscale -= (int)speedscale & ~127 ;

		for (fa=s ; fa ; fa=fa->texturechain)
			EmitSkyPolys (fa);

		glEnable (GL_BLEND);
		glBindTexture (GL_TEXTURE_2D, alphaskytexture);
		oldtexture = alphaskytexture;
		speedscale = realtime*16;
		speedscale -= (int)speedscale & ~127 ;

		for (fa=s ; fa ; fa=fa->texturechain)
			EmitSkyPolys (fa);

		glDisable (GL_BLEND);
		return;
	}

	// faster simple one-colour sky
	glBindTexture (GL_TEXTURE_2D, simpleskytexture);
	oldtexture = simpleskytexture;

	for (fa = s; fa; fa = fa->texturechain)
		EmitSimpleSkyPolys (fa);
}


//===============================================================

/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
Halflife skies are somewhat different, being just a placeholder in the map during design time,
and so are not loaded in this function.
==============
*/
void R_InitSky (gltexture_t *mt)
{
	int			i, j, p;
	byte		*src;
	unsigned	trans[128*128];
	unsigned	transpix;
	int			r, g, b;
	unsigned	*rgba;
	extern	int			skytexturenum;
	int simple_colour;

	src = (byte *)mt + mt->offsets[0];

	// make an average value for the back to avoid
	// a fringe on the top level
	r = g = b = 0;
	for (i=0 ; i<128 ; i++)
		for (j=0 ; j<128 ; j++)
		{
			p = src[i*256 + j + 128];
			rgba = &d_8to24table[p];
			trans[(i*128) + j] = *rgba;
			r += ((byte *)rgba)[0];
			g += ((byte *)rgba)[1];
			b += ((byte *)rgba)[2];
		}

	((byte *)&transpix)[0] = r/(128*128);
	((byte *)&transpix)[1] = g/(128*128);
	((byte *)&transpix)[2] = b/(128*128);
	((byte *)&transpix)[3] = 0;


	if (!solidskytexture)
		solidskytexture = texture_extension_number++;

	glBindTexture (GL_TEXTURE_2D, solidskytexture);
	oldtexture = solidskytexture;
	glTexImage2D (GL_TEXTURE_2D, 0, gl_solid_format, 128, 128, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);


	for (i=0 ; i<128 ; i++)
		for (j=0 ; j<128 ; j++)
		{
			p = src[i*256 + j];
			if (p == 0)
				trans[(i*128) + j] = transpix;
			else
				trans[(i*128) + j] = d_8to24table[p];
		}

	if (!alphaskytexture)
		alphaskytexture = texture_extension_number++;

	glBindTexture (GL_TEXTURE_2D, alphaskytexture);
	oldtexture = alphaskytexture;
	glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 128, 128, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// simpleskytexture
	switch (global_worldtype)
	{
	case 0:	// medieval
		simple_colour = 168;
		break;

	case 1:	// metal
		simple_colour = 168;
		break;

	case 2:	// base
		if (hipnotic) 
			simple_colour = 185;
		else simple_colour = 47;
		break;

	case 3: // zerstorer uses this worldtype for everything - red sky!!!
		simple_colour = 224;
		break;

	default:	// medieval by default
		simple_colour = 168;
		break;
	}

	for (i=0 ; i<128 ; i++)
		for (j=0 ; j<128 ; j++)
		{
			rgba = &d_8to24table[simple_colour];
			trans[(i*128) + j] = *rgba;
		}

	if (!simpleskytexture)
		simpleskytexture = texture_extension_number++;

	glBindTexture (GL_TEXTURE_2D, simpleskytexture);
	oldtexture = simpleskytexture;
	glTexImage2D (GL_TEXTURE_2D, 0, gl_solid_format, 128, 128, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}


/* =============================================================================================
                                         skybox stuff...
  the basic skybox loading function that originally came with the source code has been modded
  to allow for tga, jpg or pcx skyboxes. or a combination of the 3 types within the one set
  of skybox textures.
  ============================================================================================= */


/*
==================
R_LoadSkys
==================
*/

#define	SKY_TEX		2049

char	*suf[6] = {"rt", "bk", "lf", "ft", "up", "dn"};
char	*suf_hl[6] = {"r", "b", "l", "f", "u", "d"};

extern byte *image_rgba;
extern FILE *file32bit;

int LoadTGA (int *width, int *height, qboolean complain, qboolean blah);
int LoadJPG (int *width, int *height, qboolean complain, qboolean blah);
int LoadPCX (int *width, int *height, qboolean complain, qboolean use_fake_alpha);


int LoadSkyBoxPixels (char *directory, char *name, char *suffix, char *ext, int (*LoadFunc) ())
{
	char loadpath[64];
	int width = 256;
	int height = 256;
	int valid_image;

	sprintf (loadpath, "%s/%s%s.%s", directory, name, suffix, ext);
	COM_FOpenFile (loadpath, &file32bit);

	valid_image = 0;

	if (file32bit)
	{
		valid_image = (*LoadFunc) (&width, &height, false, false);

		if (valid_image)
		{
			glTexImage2D (GL_TEXTURE_2D, 0, gl_solid_format, width, height, 0, 
				GL_RGBA, GL_UNSIGNED_BYTE, image_rgba);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}

		free (image_rgba);
		fclose (file32bit);
	}

	return valid_image;
}


extern cvar_t _r_skybox;
extern qboolean hlbsp;

int R_LoadSkys (char *path)
{
	int i;

	// this gets ugly cos it has to check for skies in both /env and /gfx/env, as well as 2
	// different suffix formats, in order to support all standards. still better than forcing
	// people down one single route, though. i could make the code more elegant with a load of
	// loops checking the stuff, but would it be as readable?
	for (i = 0; i < 6; i++)
	{
		GL_Bind (SKY_TEX + i);

		// try for a TGA
		if (LoadSkyBoxPixels ("env", path, suf[i], "tga", LoadTGA)) continue;
		if (LoadSkyBoxPixels ("gfx/env", path, suf[i], "tga", LoadTGA)) continue;
		if (LoadSkyBoxPixels ("env", path, suf_hl[i], "tga", LoadTGA)) continue;
		if (LoadSkyBoxPixels ("gfx/env", path, suf_hl[i], "tga", LoadTGA)) continue;

		// try for a JPG
		if (LoadSkyBoxPixels ("env", path, suf[i], "jpg", LoadJPG)) continue;
		if (LoadSkyBoxPixels ("gfx/env", path, suf[i], "jpg", LoadJPG)) continue;
		if (LoadSkyBoxPixels ("env", path, suf_hl[i], "jpg", LoadJPG)) continue;
		if (LoadSkyBoxPixels ("gfx/env", path, suf_hl[i], "jpg", LoadJPG)) continue;

		// try for a PCX
		if (LoadSkyBoxPixels ("env", path, suf[i], "pcx", LoadPCX)) continue;
		if (LoadSkyBoxPixels ("gfx/env", path, suf[i], "pcx", LoadPCX)) continue;
		if (LoadSkyBoxPixels ("env", path, suf_hl[i], "pcx", LoadPCX)) continue;
		if (LoadSkyBoxPixels ("gfx/env", path, suf_hl[i], "pcx", LoadPCX)) continue;

		// if we've gotten this far, we have missed a skybox part, so we have no interest in
		// using this skybox... we don't set gl_useskybox to 0 or _r_skybox to "none" in this
		// case, because we might have previously loaded a valid skybox so we want to fall
		// back on that.  UPDATE - the exception is halflife BSPs which must always have a valid
		// skybox - we will fix this sometime...
		return 0;
	}

	gl_useskybox = 1;
	Cvar_Set ("_r_skybox", path);

	return 1;
}


// skybox calculation and drawing routines
// ---------------------------------------
// i'm using both types of skybox drawing in this engine (the old, slower version, and my new
// faster one) cos i need to use the older version for mirror reflections.  another reason not
// to use mirrors!
// ---------------------------------------

// multiplier for the vertices based on the current drawing distance
extern float sky_mult;

// fast vertex calculation
float QuadVecs[4][6][3] =
{
	{
		{1.0, 1.0, -1.0},
		{-1.0, 1.0, -1.0},
		{-1.0, -1.0, -1.0},
		{1.0, -1.0, -1.0},
		{1.0, 1.0, 1.0},
		{-1.0, 1.0, -1.0}
	},
	{
		{1.0, 1.0, 1.0},
		{-1.0, 1.0, 1.0},
		{-1.0, -1.0, 1.0},
		{1.0, -1.0, 1.0},
		{-1.0, 1.0, 1.0},
		{1.0, 1.0, -1.0}
	},
	{
		{1.0, -1.0, 1.0},
		{1.0, 1.0, 1.0},
		{-1.0, 1.0, 1.0},
		{-1.0, -1.0, 1.0},
		{-1.0, -1.0, 1.0},
		{1.0, -1.0, -1.0}
	},
	{
		{1.0, -1.0, -1.0},
		{1.0, 1.0, -1.0},
		{-1.0, 1.0, -1.0},
		{-1.0, -1.0, -1.0},
		{1.0, -1.0, 1.0},
		{-1.0, -1.0, -1.0}
	}
};


void MakeSkyVecs (int quad, float s, float t, int axis)
{
	vec3_t		v;

	v[0] = (QuadVecs[quad][axis][0] * sky_mult) + r_origin[0];
	v[1] = (QuadVecs[quad][axis][1] * sky_mult) + r_origin[1];
	v[2] = (QuadVecs[quad][axis][2] * sky_mult) + r_origin[2];

	glTexCoord2f (s, t);
	glVertex3fv (v);
}


int	skytexorder[6] = {0,2,1,3,4,5};


extern int gl_fogenabled;
extern glmsurface_t *skychain;
extern cvar_t r_mirroralpha;

extern cvar_t r_drawdist;
void MYgluPerspective (GLdouble fovy, GLdouble aspect, GLdouble zNear, GLdouble zFar);

extern cvar_t sky_rotate_x;
extern cvar_t sky_rotate_y;
extern cvar_t sky_rotate_z;

float rotate_x;
float rotate_y;
float rotate_z;

extern float frametime;

void R_ShutdownSkyFog (void);
void R_StartupSkyFog (void);


void R_DrawSkyBox (void)
{
	// mhquake faster skybox
	int i;
	float screenaspect;

	// increment the rotation irrespective of whether or not the sky is visible
	// in order to have it at the proper position when it becomes visible.
	rotate_x += (sky_rotate_x.value * frametime);
	if (rotate_x > 360) rotate_x -= 360; if (rotate_x < 0) rotate_x += 360;

	rotate_y += (sky_rotate_y.value * frametime);
	if (rotate_y > 360) rotate_y -= 360; if (rotate_y < 0) rotate_y += 360;

	rotate_z += (sky_rotate_z.value * frametime);
	if (rotate_z > 360) rotate_z -= 360; if (rotate_z < 0) rotate_z += 360;

	if (sky_rotate_x.value == 0) rotate_x = 0;
	if (sky_rotate_y.value == 0) rotate_y = 0;
	if (sky_rotate_z.value == 0) rotate_z = 0;

	if (!skychain) return;

	glMatrixMode (GL_PROJECTION);
	glPushMatrix ();

#ifdef MH_MIRRORS
	// enabling mirrors will disable skybox rotation, too bad, but that's life.
	// (note that if we enable mirrors, this code only gets called when there is no mirror
	// visible, so we could use rotation anyway, but this would look crap if a mirror suddenly
	// became visible and the sky snapped back to static.)
	if (r_mirroralpha.value >= 1.0)
#endif
	{
		// rotating skyboxes - currently must be set from the console
		// set up a blank identity matrix to base the rotation on (so viewangles don't interfere)
		// doesn't quite work - in fact, doesn't work at all...
		glLoadIdentity ();
		glViewport (0, 0, glwidth, glheight);
		screenaspect = (float)r_refdef.vrect.width/r_refdef.vrect.height;
		//MYgluPerspective (r_refdef.fov_y, screenaspect, 4, r_drawdist.value);
		gluPerspective (r_refdef.fov_y, screenaspect, 4, r_drawdist.value);

		// TEST - more stuff begins here
		glCullFace(GL_FRONT);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity ();

	    glRotatef (-90,  1, 0, 0);	    // put Z going up
		glRotatef (90,  0, 0, 1);	    // put Z going up

	    glRotatef (-r_refdef.viewangles[2],  1, 0, 0);
		glRotatef (-r_refdef.viewangles[0],  0, 1, 0);
		glRotatef (-r_refdef.viewangles[1],  0, 0, 1);

	    glTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);

		glDisable(GL_CULL_FACE);

		glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
		glEnable(GL_DEPTH_TEST);
		// TEST - more stuff ends here

		// rotate the skybox based on both time passed and rotation speed
		glRotatef (rotate_x, 1, 0, 0);
		glRotatef (rotate_y, 0, 1, 0);
		glRotatef (rotate_z, 0, 0, 1);

		glMatrixMode (GL_PROJECTION);
	}

	// common code for rotating and non-rotating skyboxes
	R_ShutdownSkyFog ();

	for (i = 0; i < 6; i++)
	{
		GL_Bind (SKY_TEX + skytexorder[i]);

		glBegin (GL_QUADS);

		MakeSkyVecs (0, 0.001953, 0.998047, skytexorder[i]);
		MakeSkyVecs (1, 0.001953, 0.001953, skytexorder[i]);
		MakeSkyVecs (2, 0.998047, 0.001953, skytexorder[i]);
		MakeSkyVecs (3, 0.998047, 0.998047, skytexorder[i]);

		glEnd ();
	}

	skychain = NULL;

	R_StartupSkyFog ();

	// restore the world matrix
	glPopMatrix ();
	glMatrixMode (GL_MODELVIEW);
	glLoadMatrixf (r_world_matrix);
}


#ifdef MH_MIRRORS
// classic skybox code for mirrors (slow, yuck!)

vec3_t	skyclip[6] = 
{
	{1, 1, 0},
	{1, -1, 0},
	{0, -1, 1},
	{0, 1, 1},
	{1, 0, 1},
	{-1, 0, 1} 
};


int	st_to_vec[6][3] =
{
	{3, -1, 2},
	{-3, 1, 2},
	{1, 3, 2},
	{-1, -3, 2},
	{-2,-1,3},
	{2,-1,-3}
};


int	vec_to_st[6][3] =
{
	{-2, 3, 1},
	{2, 3, -1},
	{1, 3, 2},
	{-1, 3, -2},
	{-2, -1, 3},
	{-2, 1, -3}
};


float	skymins[2][6], skymaxs[2][6];


void DrawSkyPolygon (int nump, vec3_t vecs)
{
	int		i,j;
	vec3_t	v, av;
	float	s, t, dv;
	int		axis;
	float	*vp;

	// decide which face it maps to
	VectorCopy (vec3_origin, v);

	for (i = 0, vp = vecs; i < nump; i++, vp += 3)
		VectorAdd (vp, v, v);

	av[0] = fabs(v[0]);
	av[1] = fabs(v[1]);
	av[2] = fabs(v[2]);

	if (av[0] > av[1] && av[0] > av[2])
	{
		if (v[0] < 0)
			axis = 1;
		else axis = 0;
	}
	else if (av[1] > av[2] && av[1] > av[0])
	{
		if (v[1] < 0)
			axis = 3;
		else axis = 2;
	}
	else
	{
		if (v[2] < 0)
			axis = 5;
		else axis = 4;
	}

	// project new texture coords
	for (i = 0; i < nump; i++, vecs += 3)
	{
		j = vec_to_st[axis][2];

		if (j > 0)
			dv = vecs[j - 1];
		else dv = -vecs[-j - 1];

		j = vec_to_st[axis][0];

		if (j < 0)
			s = -vecs[-j -1] / dv;
		else s = vecs[j-1] / dv;

		j = vec_to_st[axis][1];

		if (j < 0)
			t = -vecs[-j -1] / dv;
		else t = vecs[j-1] / dv;

		if (s < skymins[0][axis]) skymins[0][axis] = s;
		if (t < skymins[1][axis]) skymins[1][axis] = t;
		if (s > skymaxs[0][axis]) skymaxs[0][axis] = s;
		if (t > skymaxs[1][axis]) skymaxs[1][axis] = t;
	}
}

#define	MAX_CLIP_VERTS	64

void ClipSkyPolygon (int nump, vec3_t vecs, int stage)
{
	float	*norm;
	float	*v;
	qboolean	front, back;
	float	d, e;
	float	dists[MAX_CLIP_VERTS];
	int		sides[MAX_CLIP_VERTS];
	vec3_t	newv[2][MAX_CLIP_VERTS];
	int		newc[2];
	int		i, j;

	// can't really do anything here
	if (nump > MAX_CLIP_VERTS - 2) Sys_Error ("ClipSkyPolygon: MAX_CLIP_VERTS");

	if (stage == 6)
	{
		// fully clipped, so draw it
		DrawSkyPolygon (nump, vecs);
		return;
	}

	front = back = false;
	norm = skyclip[stage];

	for (i = 0, v = vecs; i < nump; i++, v += 3)
	{
		d = DotProduct (v, norm);

		if (d > ON_EPSILON)
		{
			front = true;
			sides[i] = SIDE_FRONT;
		}
		else if (d < ON_EPSILON)
		{
			back = true;
			sides[i] = SIDE_BACK;
		}
		else sides[i] = SIDE_ON;

		dists[i] = d;
	}

	if (!front || !back)
	{
		// not clipped
		ClipSkyPolygon (nump, vecs, stage + 1);
		return;
	}

	// clip it
	sides[i] = sides[0];
	dists[i] = dists[0];
	VectorCopy (vecs, (vecs + (i * 3)));
	newc[0] = newc[1] = 0;

	for (i = 0, v = vecs; i < nump; i++, v += 3)
	{
		switch (sides[i])
		{
		case SIDE_FRONT:
			VectorCopy (v, newv[0][newc[0]]);
			newc[0]++;
			break;

		case SIDE_BACK:
			VectorCopy (v, newv[1][newc[1]]);
			newc[1]++;
			break;

		case SIDE_ON:
			VectorCopy (v, newv[0][newc[0]]);
			newc[0]++;
			VectorCopy (v, newv[1][newc[1]]);
			newc[1]++;
			break;
		}

		if (sides[i] == SIDE_ON || sides[i + 1] == SIDE_ON || sides[i + 1] == sides[i])
			continue;

		d = dists[i] / (dists[i] - dists[i + 1]);

		for (j = 0; j < 3; j++)
		{
			e = v[j] + d * (v[j + 3] - v[j]);
			newv[0][newc[0]][j] = e;
			newv[1][newc[1]][j] = e;
		}

		newc[0]++;
		newc[1]++;
	}

	// continue
	ClipSkyPolygon (newc[0], newv[0][0], stage + 1);
	ClipSkyPolygon (newc[1], newv[1][0], stage + 1);
}
#endif


/*
=================
R_DrawSkyChain
=================
*/
void R_DrawOldSkyboxChain (glmsurface_t *s)
{
#ifdef MH_MIRRORS
	glmsurface_t	*fa;

	int		i;
	vec3_t	verts[MAX_CLIP_VERTS];
	glpoly_t	*p;

	// calculate vertex values for sky box
	for (fa = s; fa; fa = fa->texturechain)
	{
		for (p = fa->polys; p; p = p->next)
		{
			for (i = 0; i < p->numverts; i++)
				VectorSubtract (p->verts[i], r_origin, verts[i]);

			ClipSkyPolygon (p->numverts, verts[0], 0);
		}
	}
#endif
}


/*
==============
R_ClearSkyBox
==============
*/
void R_ClearSkyBox (void)
{
#ifdef MH_MIRRORS
	int		i;

	for (i = 0; i < 6; i++)
	{
		skymins[0][i] = skymins[1][i] = 9999;
		skymaxs[0][i] = skymaxs[1][i] = -9999;
	}
#endif
}


void MakeSkyVec (float s, float t, int axis)
{
#ifdef MH_MIRRORS
	vec3_t		v, b;
	int			j, k;

	b[0] = s * sky_mult;
	b[1] = t * sky_mult;
	b[2] = sky_mult;

	for (j = 0; j < 3; j++)
	{
		k = st_to_vec[axis][j];

		if (k < 0)
			v[j] = -b[-k - 1];
		else v[j] = b[k - 1];

		v[j] += r_origin[j];
	}

	// avoid bilerp seam
	s = (s + 1) * 0.5;
	t = (t + 1) * 0.5;

	if (s < 1.0 / 512) 
		s = 1.0 / 512;
	else if (s > 511.0 / 512)
		s = 511.0 / 512;

	if (t < 1.0 / 512)
		t = 1.0 / 512;
	else if (t > 511.0 / 512)
		t = 511.0 / 512;

	t = 1.0 - t;

	glTexCoord2f (s, t);
	glVertex3fv (v);
#endif
}

/*
==============
R_DrawSkyBox
==============
*/
void R_DrawOldSkyBox (void)
{
#ifdef MH_MIRRORS
	int		i;

	if (!skychain) return;
	skychain = NULL;

	R_ShutdownSkyFog ();

	for (i = 0; i < 6; i++)
	{
		if (skymins[0][i] >= skymaxs[0][i] || skymins[1][i] >= skymaxs[1][i])
			continue;

		GL_Bind (SKY_TEX + skytexorder[i]);

		glBegin (GL_QUADS);

		MakeSkyVec (skymins[0][i], skymins[1][i], i);
		MakeSkyVec (skymins[0][i], skymaxs[1][i], i);
		MakeSkyVec (skymaxs[0][i], skymaxs[1][i], i);
		MakeSkyVec (skymaxs[0][i], skymins[1][i], i);

		glEnd ();
	}

	R_StartupSkyFog ();
#endif
}
